"""
Statistics required for RCMIP analysis
"""
import numpy as np
import scipy.optimize
from scipy.stats import norm


def _diff_exp_sigma(sigma, ratio, conflevel):

    low = (1 - conflevel) / 2
    upp = 1 - low
    qno = norm.ppf(upp)
    ratio_sigma = (np.exp(qno * sigma) - 1) / (1 - np.exp(-qno * sigma))

    diff = ratio_sigma - ratio

    return diff ** 2


def _get_skewed_normal_width(ratio, conflevel):
    ratio_in = ratio if ratio >= 1 else 1 / ratio

    opt_res = scipy.optimize.minimize(
        _diff_exp_sigma,
        (0.1,),
        args=(ratio_in, conflevel),
        bounds=((10 ** -7, 10),),
        options={"maxiter": 10 ** 6},
    )

    if not opt_res.success:
        raise ValueError(
            "Optimisation failed for ratio {} and conflevel {}".format(ratio, conflevel)
        )

    skewed_normal_width = float(opt_res.x.squeeze())

    return skewed_normal_width


def get_skewed_normal(median, lower, upper, conf, input_data):
    """
    Get skewed normal distribution matching the inputs

    Parameters
    ----------
    median : float
        Median of the output distribution

    lower : float
        Lower bound of the confidence interval

    upper : float
        Upper bound of the confidence interval

    conf : float
        Confidence associated with the interval [lower, upper] e.g. 0.66
        would mean that [lower, upper] defines the 66% confidence range

    input_data : :obj:`np.ndarray`
        Points from the derived distribution to return. For each point, Y, in
        ``input_data``, we determine the value at which a cumulative
        probability of Y is achieved. As a result, all values in
        ``input_data`` must be in the range [0, 1]. Hence if you want a random
        sample from the derived skewed normal, simply make ``input_data``
        equal to a random sample of the uniform distribution [0, 1]

    Returns
    -------
    :obj:`np.ndarray`
        Points sampled from the derived skewed normal distribution based on
        ``input_data``
    """
    ratio = (upper - median) / (median - lower)
    ratio_gte_one = ratio >= 1

    skewed_normal_width = _get_skewed_normal_width(ratio, conf)

    factor_denominator = np.exp(skewed_normal_width * norm.ppf(0.5 * (1 + conf))) - 1

    if ratio_gte_one:
        factor = (upper - median) / factor_denominator
    else:
        factor = (lower - median) / factor_denominator

    shift = median - factor

    input_data = np.array(input_data)
    if ratio_gte_one:
        res_inp = input_data
    else:
        res_inp = 1 - input_data

    res = np.exp(norm.ppf(res_inp) * skewed_normal_width) * factor + shift

    return res



"""
Helpers and config for plotting
"""
import numpy as np

CLIMATE_MODEL_PALETTE = {
    "assessed range": "tab:blue",
    "MAGICC7": "tab:orange",
    "two_layer": "tab:pink",
    "Raw CMIP6 multi-model ensemble": "tab:green",
    "HadCRUT.5.0.0.0": "tab:gray",
    "HadCRUT.5.0.0.0 (GMST)": "tab:gray",
    "AR6 Prelim. FGD": "tab:gray",
    "von Shuckmann et al. 2020": "tab:purple",
}
"""dict: Colour palette used for plots coloured by climate model"""


SCENARIO_PALETTE = {
    k: np.array(v) / 256
    for k, v in {
        "ssp119": (30, 150, 132),
        "ssp126": (29, 51, 84),
        "ssp245": (234, 221, 61),
        "ssp370": (242, 17, 17),
        "ssp370-lowNTCF": (242, 17, 17),
        "ssp434": (99, 189, 229),
        "ssp460": (232, 236, 49),
        "ssp585": (132, 11, 34),
        "ssp534-over": (154, 109, 201),
    }.items()
}
"""dict: Colour palette used for plots coloured by scenario"""
SCENARIO_PALETTE["historical"] = "tab:gray"


CMIP6_NAME = "Raw CMIP6 multi-model ensemble"
"""str: String used to represent the CMIP6 multi-model ensemble in plots"""



"""
Calculation of the transient climate response (TCR)
"""
from ..database import time_mean


def calculate_transient_climate_response(
    assessed_ranges, res_calc, norm_period, evaluation_period, unit
):
    """
    Calculate the transient climate response (TCR)

    Parameters
    ----------
    assessed_ranges : :obj:`pyrcmip.assessed_ranges.AssessedRanges`
        Assessed ranges instance

    res_calc : :obj:`scmdata.ScmRun`
        Results from which the TCR is to be derived

    norm_period : list
        Years to use for normalising the data before calculating the TCR

    evaluation_period : list
        Years to use when evaluating the TCR

    unit : str
        Unit in which the TCR should be returned

    Returns
    -------
    :obj:`pd.DataFrame`
        TCR with other relevant model metadata

    Raises
    ------
    NotImplementedError
        The evaluation period is not the year 1920 i.e. the experiment does
        not match the CMIP6 specifications
    """
    res_calc_normed = assessed_ranges._get_normed_res_calc(res_calc, norm_period)

    if evaluation_period is not None and (
        len(evaluation_period) != 1 or evaluation_period[0] != 1920
    ):
        raise NotImplementedError(
            "Evaluation period other than [1920], input: {}".format(evaluation_period)
        )

    return time_mean(
        res_calc_normed.filter(
            scenario="1pctCO2",
            variable="Surface Air Temperature Change",
            unit=unit,
            region="World",
            year=range(1920, 1921),
        )
    )



"""
Handling of assessed ranges
"""
from collections import defaultdict

import matplotlib.cbook as cbook
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import tqdm.autonotebook as tqdman
from matplotlib.gridspec import GridSpec
from scmdata import run_append

from .database import time_mean
from .plotting import CLIMATE_MODEL_PALETTE, CMIP6_NAME
from .stats import get_skewed_normal


def _get_evaluation_axes():
    fig = plt.figure(figsize=(12, 7))
    gs = GridSpec(2, 1, width_ratios=[1], height_ratios=[3, 2])
    ax0 = fig.add_subplot(gs[0])
    ax1 = fig.add_subplot(gs[1], sharex=ax0)
    axes = [ax0, ax1]

    return axes


def _drop_unwanted_metadata(inp):
    out = inp.copy()
    out_meta = out.meta.columns.tolist()
    out_meta_to_drop = [
        v
        for v in out_meta
        if v
        not in [
            "model",
            "scenario",
            "region",
            "variable",
            "unit",
            "ensemble_member",
            "climate_model",
        ]
    ]
    out.drop_meta(out_meta_to_drop)

    return out


def _convert_unit(res_calc, unit):
    # TODO: check if this is still needed...
    # some scmdata bug requires this workaround...
    units_to_convert = [u for u in res_calc.get_unique_meta("unit") if u != unit]
    if units_to_convert:
        print("Converting {} to {}".format(units_to_convert, unit))
        res_calc_leave = _drop_unwanted_metadata(
            res_calc.filter(unit=unit, log_if_empty=False)
        )
        res_calc_converted = _drop_unwanted_metadata(
            res_calc.filter(unit=unit, keep=False, log_if_empty=False).convert_unit(
                unit
            )
        )

        res_calc = run_append([res_calc_leave, res_calc_converted])

    res_calc = _drop_unwanted_metadata(res_calc)

    return res_calc


class AssessedRanges:
    """
    Class for handling assessed ranges and performing operations with them.

    For example, getting values for specific metrics and plotting results against
    assessed ranges.
    """

    def __init__(self, db):
        """
        Initialise

        Parameters
        ----------
        db : :obj:`pd.DataFrame`
            Assessed ranges data
        """
        self.db = db.copy()

    def head(self, n=5):
        """
        Show head of ``self.db``

        Parameters
        ----------
        n : int
            Number of rows to show

        Returns
        -------
        :obj:`pd.DataFrame`
            Head of ``self.db``
        """
        return self.db.head(n)

    def tail(self, n=5):
        """
        Show tail of ``self.db``

        Parameters
        ----------
        n : int
            Number of rows to show

        Returns
        -------
        :obj:`pd.DataFrame`
            Tail of ``self.db``
        """
        return self.db.tail(n)

    def get_col_for_metric(self, metric, col, single_value=True):
        """
        TODO: docstring
        """
        out = self.db.loc[self.db["RCMIP name"] == metric, col].tolist()

        assert len(out) == 1, "{} {}: {}".format(metric, col, out)

        if not single_value:
            return [v.strip() for v in out[0].split(",")]

        return out[0]

    def get_variables_regions_scenarios_for_metric(self, metric, single_value=True):
        """
        TODO: docstring
        """
        return [
            self.get_col_for_metric(metric, "RCMIP {}".format(ids), single_value=False)
            for ids in ("variable", "region", "scenario")
        ]

    def get_norm_period_evaluation_period(self, metric):
        """
        TODO: docstring
        """
        norm_period_start = self.get_col_for_metric(metric, "norm_period_start")

        norm_period_end = self.get_col_for_metric(metric, "norm_period_end")

        if not np.isnan(norm_period_start) and not np.isnan(norm_period_end):
            norm_period = range(int(norm_period_start), int(norm_period_end) + 1)
        else:
            norm_period = None

        evaluation_period_start = self.get_col_for_metric(
            metric, "evaluation_period_start"
        )

        evaluation_period_end = self.get_col_for_metric(metric, "evaluation_period_end")

        if not np.isnan(evaluation_period_start) and not np.isnan(
            evaluation_period_end
        ):
            evaluation_period = range(
                int(evaluation_period_start), int(evaluation_period_end) + 1
            )
        else:
            evaluation_period = None

        return norm_period, evaluation_period

    def check_norm_period_evaluation_period_against_data(
        self, norm_period, evaluation_period, res_calc
    ):
        """
        TODO: docstring
        """
        res_min_year = res_calc["year"].min()
        res_max_year = res_calc["year"].max()

        if norm_period is not None:
            if res_min_year > min(norm_period):
                raise ValueError(
                    "Res min. year: {}, greater than ref. period min.: {}".format(
                        res_min_year, min(norm_period)
                    )
                )

            if res_max_year < max(norm_period):
                raise ValueError(
                    "Res max. year: {}, less than ref. period max.: {}".format(
                        res_max_year, max(norm_period)
                    )
                )

        if evaluation_period is not None:
            if res_min_year > min(evaluation_period):
                raise ValueError(
                    "Res min. year: {}, greater than evalution period min.: {}".format(
                        res_min_year, min(evaluation_period)
                    )
                )

            if res_max_year < max(evaluation_period):
                raise ValueError(
                    "Res max. year: {}, less than evalution period max.: {}".format(
                        res_max_year, max(evaluation_period)
                    )
                )

    def get_assessed_range_for_boxplot(self, metric):
        """
        TODO: docstring
        """
        value_cols = [
            "very_likely__lower",
            "likely__lower",
            "central",
            "likely__upper",
            "very_likely__upper",
        ]
        values = {
            k: float(self.db.loc[self.db["RCMIP name"] == metric, k].values.squeeze())
            for k in value_cols
        }

        if np.isnan(values["very_likely__lower"]):
            lower = values["likely__lower"]
            upper = values["likely__upper"]
            conf = 0.66
        else:
            lower = values["very_likely__lower"]
            upper = values["very_likely__upper"]
            conf = 0.9

        if np.isnan(values["central"]):
            out = np.zeros(100) * np.nan
        else:
            # generate array which will plot as desired
            out = get_skewed_normal(
                values["central"], lower, upper, conf, np.random.random(2 * 10 ** 4)
            )

        ar_for_boxplot = pd.DataFrame(out, columns=[metric])
        ar_for_boxplot["Source"] = "assessed range"
        ar_for_boxplot["unit"] = self.get_col_for_metric(
            metric, "unit", single_value=True
        )

        return ar_for_boxplot

    def plot_metric_and_results(self, metric, model_results, axes=None):
        """
        TODO: docstring
        """
        if axes is None:
            axes = _get_evaluation_axes()
        else:
            assert len(axes) == 2

        return self._plot(metric, model_results, axs=axes, box_only=False)

    def plot_metric_and_results_box_only(self, metric, model_results, ax=None):
        """
        TODO: docstring
        """
        if ax is None:
            ax = plt.figure().add_subplot(111)

        return self._plot(metric, model_results, axs=ax, box_only=True)

    def _plot(self, metric, model_results, axs, box_only):
        res = model_results.pivot_table(
            values="value",
            columns="RCMIP name",
            index=list(set(model_results.columns) - {"value", "RCMIP name"}),
        ).reset_index()
        res["Source"] = res["climate_model"]
        res["Source"] = res["Source"].apply(
            lambda x: CMIP6_NAME if x.startswith("CMIP6") else x
        )

        ar = self.get_assessed_range_for_boxplot(metric)

        box_plot_df = pd.concat([ar, res])[[metric, "Source", "unit"]]

        unit = box_plot_df["unit"].unique().tolist()
        assert len(unit) == 1, unit

        if box_only:
            box_ax = axs
        else:
            box_ax = axs[1]

        stats = {}
        for source, df in box_plot_df.groupby("Source"):
            data = df[metric].values.squeeze()
            stats[source] = cbook.boxplot_stats(data, labels=[source])[0]
            if source == "SOD assessed range":
                values = self.get_assessed_range_values_for_metric(metric)

                if not np.isnan(values["central"]):
                    stats[source]["med"] = values["central"]
                else:
                    stats[source]["med"] = np.percentile(data, 50)

                if not np.isnan(values["likely__lower"]):
                    stats[source]["q1"], stats[source]["q3"] = (
                        values["likely__lower"],
                        values["likely__upper"],
                    )
                else:
                    stats[source]["q1"], stats[source]["q3"] = np.percentile(
                        data, [17, 83]
                    )

                if not np.isnan(values["very_likely__lower"]):
                    stats[source]["whislo"], stats[source]["whishi"] = (
                        values["very_likely__lower"],
                        values["very_likely__upper"],
                    )
                else:
                    stats[source]["whislo"], stats[source]["whishi"] = np.percentile(
                        data, [5, 95]
                    )

            else:
                stats[source]["q1"], stats[source]["q3"] = np.percentile(data, [17, 83])
                stats[source]["whislo"], stats[source]["whishi"] = np.percentile(
                    data, [5, 95]
                )

        box_ax.bxp(
            list(stats.values()),
            showfliers=False,
            vert=box_only,
            #             rot=90 if box_only else 0,
            #             positions
        )
        box_ax.set_xlabel(unit[0])

        xlim = box_ax.get_xlim()

        if box_only:
            box_ax.set_title(metric)
            box_ax.grid(axis="y")
        else:
            box_ax.set_title("")
            box_ax.grid(axis="x")

        if not box_only:
            pdf_ax = axs[0]
            for label, df in box_plot_df.groupby("Source"):
                sns.distplot(
                    df[metric],
                    bins=50,
                    norm_hist=True,
                    label=label,
                    ax=pdf_ax,
                    color=CLIMATE_MODEL_PALETTE[label],
                )

            pdf_ax.legend()
            pdf_ax.set_title(metric)
            pdf_ax.set_xlim(xlim)
            pdf_ax.set_xlabel("")

        plt.suptitle("")

        return axs

    def get_results_summary_table_for_metric(self, metric, model_results):
        """
        TODO: docstring
        """
        summary_table = defaultdict(list)
        assessed_range_quantiles = {
            "very_likely__lower": 0.05,
            "likely__lower": 0.17,
            "central": 0.5,
            "likely__upper": 0.83,
            "very_likely__upper": 0.95,
        }
        for label, quantile in tqdman.tqdm(
            assessed_range_quantiles.items(), leave=False
        ):
            assessed_value = self.get_col_for_metric(metric, label)
            if np.isnan(assessed_value):
                continue

            for climate_model, df in model_results.groupby("climate_model"):
                model_quantile = df["value"].quantile(quantile)

                summary_table["climate_model"].append(climate_model)
                summary_table["assessed_range_label"].append(label)
                summary_table["assessed_range_value"].append(assessed_value)
                summary_table["climate_model_value"].append(model_quantile)
                summary_table["percentage_difference"].append(
                    (model_quantile - assessed_value) / assessed_value * 100
                )

        summary_table = pd.DataFrame(summary_table)
        summary_table["metric"] = metric
        summary_table["unit"] = self.get_col_for_metric(metric, "unit")

        return summary_table

    def plot_against_results(
        self, results_database, custom_calculations={}, exclude_models=None
    ):
        """
        TODO: docstring
        """
        summary_table = []
        for metric in tqdman.tqdm(self.db["RCMIP name"].tolist()):
            eval_method = self.get_col_for_metric(metric, "RCMIP evaluation method")
            if isinstance(eval_method, float) and np.isnan(eval_method):
                print("No evaluation method for {}".format(metric))
                continue

            (
                variables,
                regions,
                scenarios,
            ) = self.get_variables_regions_scenarios_for_metric(
                metric, single_value=False
            )

            try:
                res_calc = run_append(
                    [
                        results_database.load_data(variable=v, region=r, scenario=s)
                        for v in variables
                        for r in regions
                        for s in scenarios
                    ]
                ).time_mean("AC")
            except IndexError:
                print("No data for: {}".format(metric))
                continue

            res_calc = _drop_unwanted_metadata(res_calc)
            if exclude_models is not None:
                res_calc = res_calc.filter(climate_model=exclude_models, keep=False)

            derived = self.calculate_metric_from_results(
                metric, res_calc, custom_calculations=custom_calculations
            )

            summary_table_metric = self._plot_pdf_and_box_and_get_summary_table(
                metric, derived
            )

            summary_table.append(summary_table_metric)

        summary_table = pd.concat(summary_table).reset_index(drop=True)

        return summary_table

    def plot_model_reported_against_results(self, model_reported):
        """
        TODO: docstring
        """
        summary_table = []

        for label, df in model_reported.groupby("RCMIP name"):
            summary_table_metric = self._plot_pdf_and_box_and_get_summary_table(
                label, df
            )

            summary_table.append(summary_table_metric)

        summary_table = pd.concat(summary_table).reset_index(drop=True)

        return summary_table

    def _plot_pdf_and_box_and_get_summary_table(self, metric, derived):
        axes = _get_evaluation_axes()

        axes = self.plot_metric_and_results(metric, derived, axes=axes)
        plt.tight_layout()

        ax = plt.figure(figsize=(6, 4)).add_subplot(111)
        self.plot_metric_and_results_box_only(metric, derived, ax=ax)

        plt.tight_layout()
        plt.show()

        summary_table_metric = self.get_results_summary_table_for_metric(
            metric, derived
        )

        return summary_table_metric

    def _get_normed_res_calc(self, res_calc, norm_period):
        if norm_period is not None:
            res_calc_normed = res_calc.relative_to_ref_period_mean(year=norm_period)
        else:
            res_calc_normed = res_calc

        return res_calc_normed

    def calculate_metric_from_results(self, metric, res_calc, custom_calculations={}):
        """
        TODO: docstring
        """
        eval_method = self.get_col_for_metric(metric, "RCMIP evaluation method")

        (norm_period, evaluation_period) = self.get_norm_period_evaluation_period(
            metric
        )
        self.check_norm_period_evaluation_period_against_data(
            norm_period, evaluation_period, res_calc
        )

        unit = self.get_col_for_metric(metric, "unit")

        if eval_method == "mean":
            print(norm_period)
            res_calc_normed = self._get_normed_res_calc(res_calc, norm_period)
            res_calc_normed = _convert_unit(res_calc_normed, unit)

            derived = time_mean(res_calc_normed.filter(year=evaluation_period))

            if metric.startswith("Minus"):
                derived *= -1

        elif eval_method == "integral":
            derived = self._calculate_integral(
                res_calc, norm_period, evaluation_period, unit
            )

        elif eval_method == "custom":
            try:
                calc = [c for f, c in custom_calculations.items() if f(metric)][0]

                derived = calc(self, res_calc, norm_period, evaluation_period, unit)

            except IndexError:
                raise IndexError(
                    "No custom method implementation for {}".format(metric)
                )

        else:
            raise NotImplementedError(eval_method)

        derived = derived.to_frame().reset_index()
        derived["RCMIP name"] = metric

        return derived



"""
Input and output handling
"""
import contextlib
import gzip
import io
import logging
import os
import os.path
import tempfile

import pandas as pd
from scmdata.run import ScmRun, run_append

LOGGER = logging.getLogger(__name__)


def ensure_dir_exists(fp):
    """
    Ensure directory exists

    Parameters
    ----------
    fp : str
        Filepath of which to ensure the directory exists
    """
    dir_to_check = os.path.dirname(fp)
    if not os.path.isdir(dir_to_check):
        LOGGER.info("Creating {}".format(dir_to_check))
        os.makedirs(dir_to_check)


def read_results_submission(results):
    """
    Read results submission

    Parameters
    ----------
    results : str or list of str
        Files to read in. All files to be read should be formatted as csv or
        xlsx files following the formatting defined in the template
        spreadsheet.

    Returns
    -------
    :obj:`scmdata.ScmRun`
        Results read in from the submission(s)
    """
    if isinstance(results, str):
        results = [results]

    db = []
    for rf in results:
        LOGGER.info("Reading %s", rf)

        if rf.endswith(".nc"):
            LOGGER.info("Assuming netCDF format")
            loaded = ScmRun.from_nc(rf)
        else:
            if rf.endswith(".xlsx") or rf.endswith(".xls"):
                LOGGER.info("Assuming excel format")
                loaded = pd.read_excel(rf, sheet_name="your_data")
            else:
                LOGGER.info("Assuming csv format")
                loaded = pd.read_csv(rf)

            LOGGER.debug("Converting all columns to lowercase")
            loaded.columns = loaded.columns.astype(str).str.lower()
            loaded = ScmRun(loaded)

        db.append(ScmRun(loaded))

    LOGGER.info("Joining results together")
    db = run_append(db)

    return db


def read_submission_model_reported(fp):
    """
    Read the model reported component of a submission

    Parameters
    ----------
    fp : str
        Filepath to read

    Returns
    -------
    :obj:`pd.DataFrame`
    """
    out = pd.read_csv(fp)

    return out


def read_submission_model_metadata(fp):
    """
    Read the model metadata component of a submission

    Parameters
    ----------
    fp : str
        Filepath to read

    Returns
    -------
    :obj:`pd.DataFrame`
    """
    out = pd.read_csv(fp)

    if out.columns[0] == "Unnamed: 0":
        # assume data was saved directly from excel sheet
        # drop first column
        out = out.iloc[:, 1:]
        column_map = {
            "ClimateModel": "climate_model",
            "Climate Model Name": "climate_model_name",
            "Climate Model Version": "climate_model_version",
            "Climate Model Configuration Label": "climate_model_configuration_label",
            "Model Configuration Description": "climate_model_configuration_description",
            "Project": "project",
            "Name of Person": "name_of_person",
            "Literature Reference": "literature_reference",
        }
        out.columns = out.columns.map(column_map)

    return out


@contextlib.contextmanager
def temporary_file_to_upload(df, max_size=1024, compress=False):
    """
    Create a gzipped temporary serialized version of a file to upload

    Attempts to keep the file in memory until it exceeds `max_size`. The file is then stored on-disk
    and cleaned up at the end of the context.

    The temporary location can be overriden using the `TMPDIR` environment variable as per
    https://docs.python.org/3/library/tempfile.html#tempfile.gettempdir

    Parameters
    ----------
    df : :obj:`scmdata.ScmRun` or :obj:`pd.DataFrame`
        Run to store

    max_size: int or float
        Max size in MB before file is temporarily streamed to disk. Defaults to 1GB

    Returns
    -------
    :obj:`tempfile.SpooledTemporaryFile`
        Open file object ready to be streamed
    """
    if isinstance(df, ScmRun):
        df = df.timeseries().reset_index()

    if compress:
        buffer = tempfile.SpooledTemporaryFile(max_size=max_size * 1024 * 1024)
        with gzip.GzipFile(mode="w", fileobj=buffer) as gz_file:
            df.to_csv(io.TextIOWrapper(gz_file, "utf8"), index=False)
    else:
        buffer = io.BytesIO(df.to_csv(index=False).encode("utf8"))

    buffer.seek(0)

    try:
        yield buffer
    finally:
        buffer.close()



"""
Utility functions for validating RCMIP submissions
"""
import os.path

import pandas as pd
import pkg_resources

SUBMISSION_TEMPLATE = os.path.join(
    "..", "data", "rcmip-data-submission-template-v5-0-0.xlsx"
)


def load_submission_template_definitions(fp=None, component="variables"):
    """
    Load submission template definitions

    Parameters
    ----------
    fp : str
        Filepath from which to load the definitions. If ``None``, definitions
        will be loaded from
        ``pyrcmip/data/rcmip-data-submission-template-v4-0-0.xlsx``.

    component : {"variables", "regions", "scenarios"}
        Definitions section to load

    Returns
    -------
    :obj:`pd.DataFrame`
    """
    if fp is None:
        fp = pkg_resources.resource_stream(__name__, SUBMISSION_TEMPLATE)

    if component == "variables":
        component_kwargs = {
            "sheet_name": "variable_definitions",
            "usecols": range(1, 6),
        }

    elif component == "regions":
        component_kwargs = {
            "sheet_name": "region_definitions",
        }

    elif component == "scenarios":
        component_kwargs = {
            "sheet_name": "scenario_info",
            "skiprows": 2,
            "usecols": range(1, 5),
        }

    else:
        raise NotImplementedError("Unrecognised component: {}".format(component))

    out = pd.read_excel(fp, **component_kwargs)
    out.columns = out.columns.str.lower()

    if component == "variables":
        out["tier"] = out["tier"].astype(int)

    elif component == "scenarios":
        out = out.dropna()
        column_map = {
            "# scenario id": "scenario_id",
            "# scenario description": "description",
            "#scenario specification": "specification",
            "# tier in rcmp": "tier",
        }
        out.columns = out.columns.map(column_map)
        out["tier"] = out["tier"].astype(int)

    return out



"""
Console script for rcmip.
"""
import logging
import sys
from os.path import join

import boto3
import click
import jwt
import semver
from botocore.exceptions import ClientError

from pyrcmip.io import (
    read_results_submission,
    read_submission_model_metadata,
    read_submission_model_reported,
    temporary_file_to_upload,
)
from pyrcmip.validate import validate_submission_bundle

LOGGER = logging.getLogger(__name__)
DEFAULT_LOG_FORMAT = "{process} {asctime} {levelname}:{name}:{message}"


class ColorFormatter(logging.Formatter):
    """
    Colour formatter for log messages

    A handy little tool for making our log messages look slightly prettier
    """

    colors = {
        "DEBUG": dict(fg="blue"),
        "INFO": dict(fg="green"),
        "WARNING": dict(fg="yellow"),
        "Error": dict(fg="red"),
        "ERROR": dict(fg="red"),
        "EXCEPTION": dict(fg="red"),
        "CRITICAL": dict(fg="red"),
    }

    def format(self, record):
        """
        Format a record so it has pretty colours

        Parameters
        ----------
        record : :obj:`logging.LogRecord`
            Record to format

        Returns
        -------
        str
            Formatted message string
        """
        formatted_message = super(ColorFormatter, self).format(record)

        level = record.levelname

        if level in self.colors:
            level_colour = click.style("{}".format(level), **self.colors[level])
            formatted_message = formatted_message.replace(level, level_colour)

        return formatted_message


class ClickHandler(logging.Handler):
    """
    Handler which emits using click when going to stdout
    """

    _use_stderr = True

    def emit(self, record):
        """
        Emit a record

        Parameters
        ----------
        record : :obj:`logging.LogRecord`
            Record to emit
        """
        try:
            msg = self.format(record)
            click.echo(msg, err=self._use_stderr)

        except Exception:  # pragma: no cover
            self.handleError(record)


_default_handler = ClickHandler()
_default_handler.formatter = ColorFormatter(DEFAULT_LOG_FORMAT, style="{")


@click.group(name="rcmip")
@click.option(
    "--log-level",
    default="INFO",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "EXCEPTION", "CRITICAL"]),
)
def cli(log_level):
    """
    Command-line interface for pyrcmip
    """
    root = logging.getLogger()
    root.handlers.append(_default_handler)
    root.setLevel(log_level)

    logging.captureWarnings(True)


def _read_bundle(timeseries, model_reported, metadata):
    try:
        scmrun = read_results_submission(timeseries)
    except Exception as e:
        LOGGER.exception("reading timeseries failed")
        raise click.ClickException(str(e))

    try:
        model_reported_df = read_submission_model_reported(model_reported)
    except Exception as e:
        LOGGER.exception("reading model_reported failed")
        raise click.ClickException(str(e))

    try:
        metadata_df = read_submission_model_metadata(metadata)
    except Exception as e:
        LOGGER.exception("reading metadata failed")
        raise click.ClickException(str(e))

    return scmrun, model_reported_df, metadata_df


timeseries = click.argument(
    "timeseries", nargs=-1, required=True, type=click.Path(exists=True, dir_okay=False)
)
model_reported = click.argument(
    "model_reported",
    nargs=1,
    required=True,
    type=click.Path(exists=True, dir_okay=False),
)
metadata = click.argument(
    "metadata", nargs=1, required=True, type=click.Path(exists=True, dir_okay=False)
)


@cli.command()
@timeseries
@model_reported
@metadata
def validate(timeseries, model_reported, metadata):
    """
    Validate submission input

    Three different types of input data are required for validation, namely:

    One or more ``TIMESERIES`` files in which the timeseries output is stored. These should be
    CSV or NetCDF files conforming to the format expected by ``scmdata``. Multiple
    timeseries inputs can be specified, but care must be taken to ensure that all of
    the individual timeseries have unique metadata.

    ``MODEL_REPORTED`` is the CSV file in which the model reported metrics are stored.

    ``METADATA`` is the CSV file in which the metadata output is stored.
    """
    scmrun, model_reported_df, metadata_df = _read_bundle(
        timeseries, model_reported, metadata
    )

    try:
        validate_submission_bundle(scmrun, model_reported_df, metadata_df)
    except Exception as e:
        raise click.ClickException(str(e))


def validate_version(ctx, param, value):
    """
    Validate version string

    Parameters
    ----------
    ctx
        Not used

    param
        Not used

    value : str
        Version string to validate

    Returns
    -------
    str
        Validated version string

    Raises
    ------
    :obj:`click.BadParameter`
        Version string cannot be passed or does not follow semantic versioning
    """
    try:
        s = semver.VersionInfo.parse(value)

        if s.prerelease is None and s.build is None:
            return value
        else:
            raise click.BadParameter(
                "Version must only contain major, minor and patch values"
            )
    except ValueError:
        raise click.BadParameter("Cannot parse version string")


@cli.command()
@click.option(
    "--token",
    required=True,
    help="Authentication token. Contact zebedee.nicholls@climate-energy-college.org for a token",
)
@click.option("--bucket", default="rcmip-uploads-au")
@click.option("--model", required=True)
@click.option(
    "--version",
    required=True,
    callback=validate_version,
    help="Version of the data being uploaded. Must be a valid semver version string (https://semver.org/). "
    "For example 2.0.0",
)
@timeseries
@model_reported
@metadata
def upload(token, bucket, model, version, timeseries, model_reported, metadata):
    """
    Validate and upload data to RCMIP's S3 bucket.

    All the files for a given version have to be uploaded together.

    One or more ``TIMESERIES`` files in which the timeseries output is stored. These should be
    CSV or NetCDF files conforming to the format expected by ``scmdata``. Multiple
    timeseries inputs can be specified, but care must be taken to ensure that all of
    the individual timeseries have unique metadata.

    ``MODEL_REPORTED`` is the CSV file in which the model reported metrics are stored.

    ``METADATA`` is the CSV file in which the metadata output is stored.
    """
    try:
        scmrun, model_reported_df, metadata_df = _read_bundle(
            timeseries, model_reported, metadata
        )
        (
            scmrun_rcmip_compatible,
            model_reported_rcmip_compatible,
            metadata_rcmip_compatiable,
        ) = validate_submission_bundle(scmrun, model_reported_df, metadata_df)
    except Exception:
        raise click.ClickException("Validation failed. Fix issues and rerun")

    # Upload data to S3
    t = jwt.decode(token, verify=False)
    session = boto3.session.Session(
        aws_access_key_id=t["access_key_id"],
        aws_secret_access_key=t["secret_access_key"],
    )
    client = session.client("s3")

    root_key = "{}/{}/{}".format(t["org"], model, version)

    # Check if this version is already uploaded (using the {key}-complete dummy file)
    try:
        LOGGER.debug("Checking if object with key {} exists".format(root_key))
        client.head_object(Bucket=bucket, Key=root_key + "-complete")

        raise click.ClickException(
            "Data for this version has already been uploaded. Increment the version and try again"
        )
    except ClientError:
        LOGGER.debug("Object with key {} does not exist".format(root_key))

    for name, obj, compress in [
        ("data", scmrun_rcmip_compatible, True),
        ("model_reported", model_reported_rcmip_compatible, False),
        ("metadata", metadata_rcmip_compatiable, False),
    ]:
        try:
            # Small files are kept in memory while bigger files are written to disk and automatically cleaned up
            LOGGER.info("Preparing {} for upload".format(name))
            with temporary_file_to_upload(obj, max_size=1024, compress=compress) as fh:
                try:
                    fname = "rcmip-{}-{}-{}.csv".format(model, version, name)
                    if compress:
                        fname += ".gz"
                    key = join(root_key, fname)
                    LOGGER.info("Uploading {}".format(fname))
                    client.upload_fileobj(fh, Bucket=bucket, Key=key)
                except ClientError:  # pragma: no cover
                    LOGGER.exception("Failed to upload file")
                    raise click.ClickException("Failed to upload file")
        except Exception as e:
            LOGGER.exception(e)

    # Finally mark the upload as complete by uploading a dummy file
    # Writing this dummy file will be used to start the processing of the upload
    client.put_object(Bucket=bucket, Key=root_key + "-complete")

    LOGGER.info("All files uploaded successfully")


def run_cli():
    """
    Run command-line interface

    TODO: fix this so environment variables can be used
    """
    sys.exit(cli(auto_envvar_prefix="RCMIP"))  # pragma: no cover


if __name__ == "__main__":
    run_cli()  # pragma: no cover
